@@ -2,7 +2,7 @@
  * pstree.c - display process tree
  *
  * Copyright (C) 1993-2002 Werner Almesberger
- * Copyright (C) 2002-2009 Craig Small
+ * Copyright (C) 2002-2017 Craig Small
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,18 +41,27 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <limits.h>
 
 #include "i18n.h"
 #include "comm.h"
 
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
+#else
+typedef void* security_context_t; /* DUMMY to remove most ifdefs */
 #endif                                /*WITH_SELINUX */
 
 extern const char *__progname;
 
 #define PROC_BASE    "/proc"
 
+#if defined(__FreeBSD_kernel__) || defined(__FreeBSD__)
+#define ROOT_PID 0
+#else
+#define ROOT_PID 1
+#endif /* __FreeBSD__ */
+
 /* UTF-8 defines by Johan Myreen, updated by Ben Winslow */
 #define UTF_V        "\342\224\202"        /* U+2502, Vertical line drawing char */
 #define UTF_VR        "\342\224\234"        /* U+251C, Vertical and right */
@@ -68,15 +77,38 @@ extern const char *__progname;
 #define VT_UR        "m"
 #define        VT_HD        "w"
 
+#define THREAD_FORMAT        "{%.*s}"        /* Format for thread names */
+
+enum ns_type {
+    CGROUPNS = 0,
+    IPCNS,
+    MNTNS,
+    NETNS,
+    PIDNS,
+    USERNS,
+    UTSNS,
+    NUM_NS
+};
+
+static const char *ns_names[] = {
+    [CGROUPNS] = "cgroup",
+    [IPCNS] = "ipc",
+    [MNTNS] = "mnt",
+    [NETNS] = "net",
+    [PIDNS] = "pid",
+    [USERNS] = "user",
+    [UTSNS] = "uts",
+};
+
 typedef struct _proc {
-    char comm[COMM_LEN + 1];
+    char comm[COMM_LEN + 2 + 1]; /* add another 2 for thread brackets */
     char **argv;                /* only used : argv[0] is 1st arg; undef if argc < 1 */
     int argc;                        /* with -a   : number of arguments, -1 if swapped    */
     pid_t pid;
+    pid_t pgid;
     uid_t uid;
-#ifdef WITH_SELINUX
     security_context_t scontext;
-#endif                                /*WITH_SELINUX */
+    ino_t ns[NUM_NS];
     char flags;
     struct _child *children;
     struct _proc *parent;
@@ -92,6 +124,12 @@ typedef struct _child {
     struct _child *next;
 } CHILD;
 
+struct ns_entry {
+    ino_t number;
+    CHILD *children;
+    struct ns_entry *next;
+};
+
 static struct {
     const char *empty_2;        /*    */
     const char *branch_2;        /* |- */
@@ -122,17 +160,159 @@ static int capacity = 0;
 static int *width = NULL;
 static int *more = NULL;
 
-static int print_args = 0, compact = 1, user_change = 0, pids = 0, by_pid =
-    0, trunc = 1, wait_end = 0;
-#ifdef WITH_SELINUX
+static int print_args = 0, compact = 1, user_change = 0, pids = 0, pgids = 0,
+    show_parents = 0, by_pid = 0, trunc = 1, wait_end = 0, ns_change = 0,
+    thread_names = 0, hide_threads = 0;
 static int show_scontext = 0;
-#endif                                /*WITH_SELINUX */
 static int output_width = 132;
 static int cur_x = 1;
 static char last_char = 0;
 static int dumped = 0;                /* used by dump_by_user */
 static int charlen = 0;                /* length of character */
 
+const char *get_ns_name(enum ns_type id) {
+    if (id >= NUM_NS)
+        return NULL;
+    return ns_names[id];
+}
+
+static enum ns_type get_ns_id(const char *name) {
+    int i;
+
+    for (i = 0; i < NUM_NS; i++)
+        if (!strcmp(ns_names[i], name))
+            return i;
+    return NUM_NS;
+}
+
+static int verify_ns(enum ns_type id)
+{
+    char filename[50];
+    struct stat s;
+
+    snprintf(filename, 50, "/proc/%i/ns/%s", getpid(), get_ns_name(id));
+
+    return stat(filename, &s);
+}
+
+static inline void new_proc_ns(PROC *ns_task)
+{
+    struct stat st;
+    char buff[50];
+    pid_t pid = ns_task->pid;
+    int i;
+
+    for (i = 0; i < NUM_NS; i++) {
+        snprintf(buff, sizeof(buff), "/proc/%i/ns/%s", pid,
+                 get_ns_name(i));
+        if (stat(buff, &st)) {
+            ns_task->ns[i] = 0;
+            continue;
+        }
+        ns_task->ns[i] = st.st_ino;
+    }
+}
+
+static void find_ns_and_add(struct ns_entry **root, PROC *r, enum ns_type id)
+{
+    struct ns_entry *ptr, *last = NULL;
+    CHILD *tmp_child, **c;
+
+    for (ptr = *root; ptr; ptr = ptr->next) {
+        if (ptr->number == r->ns[id])
+            break;
+        last = ptr;
+    }
+
+    if (!ptr) {
+
+        if (!(ptr = malloc(sizeof(*ptr)))) {
+            perror("malloc");
+            exit(1);
+        }
+
+        memset(ptr, 0, sizeof(*ptr));
+        ptr->number = r->ns[id];
+        if (*root == NULL)
+            *root = ptr;
+        else
+            last->next = ptr;
+    }
+
+    /* move the child to under the namespace's umbrella */
+    for (c = &ptr->children; *c; c = &(*c)->next)
+        ;
+
+    if (!(*c = malloc(sizeof(CHILD)))) {
+        perror("malloc");
+        exit(1);
+    }
+
+    (*c)->child = r;
+    (*c)->next = NULL;
+
+    /* detaching from parent */
+    if (r->parent) {
+        for (c = &r->parent->children; *c; c = &(*c)->next) {
+            if ((*c)->child == r) {
+                tmp_child = (*c)->next;
+                free(*c);
+                *c = tmp_child;
+                break;
+            }
+        }
+        r->parent = NULL;
+    }
+
+}
+
+static PROC *find_proc(pid_t pid);
+static void sort_by_namespace(PROC *r, enum ns_type id, struct ns_entry **root)
+{
+    CHILD *walk, *next;
+
+    /* first run, find the first process */
+    if (!r) {
+        r = find_proc(1);
+        if (!r)
+            return;
+    }
+
+    if (r->parent == NULL || r->parent->ns[id] != r->ns[id])
+        find_ns_and_add(root, r, id);
+
+    walk = r->children;
+    while (walk) {
+        next = walk->next;
+        sort_by_namespace(walk->child, id, root);
+        walk = next;
+    }
+}
+
+static void fix_orphans(security_context_t scontext);
+
+/*
+ * Determine the correct output width, what we use is:
+ */
+static int get_output_width(void)
+{
+    char *ep, *env_columns;
+    struct winsize winsz;
+
+    env_columns = getenv("COLUMNS");
+    if (env_columns && *env_columns) {
+        long t;
+        t = strtol(env_columns, &ep, 0);
+        if (!*ep && (t > 0) && (t < 0x7fffffffL))
+            return (int)t;
+    }
+    if (ioctl(1, TIOCGWINSZ, &winsz) >= 0)
+        if (winsz.ws_col)
+            return winsz.ws_col;
+    return 132;
+
+}
+
 /*
  * Allocates additional buffer space for width and more as needed.
  * The first call will allocate the first buffer.
@@ -174,6 +354,48 @@ static void free_buffers()
     capacity = 0;
 }
 
+static void free_children(CHILD *children)
+{
+    CHILD *walk, *next;
+
+    walk = children;
+    while (walk != NULL) {
+        next = walk->next;
+        free(walk);
+        walk = next;
+    }
+}
+
+static void free_proc()
+{
+    PROC *walk, *next;
+    walk = list;
+    while (walk != NULL) {
+        next = walk->next;
+        free_children(walk->children);
+        if (walk->argv) {
+            free(walk->argv[0]);
+            free(walk->argv);
+        }
+        free(walk);
+        walk = next;
+    }
+    list = NULL;
+}
+
+static void free_namespace(struct ns_entry **nsroot)
+{
+    struct ns_entry *walk, *next;
+    walk = *nsroot;
+    while (walk != NULL) {
+        next = walk->next;
+        free_children(walk->children);
+        free(walk);
+        walk = next;
+    }
+    *nsroot = NULL;
+}
+
 static void out_char(char c)
 {
     if (charlen == 0) {                /* "new" character */
@@ -221,15 +443,12 @@ static int out_int(int x)
     return digits;
 }
 
-#ifdef WITH_SELINUX
 static void out_scontext(security_context_t scontext)
 {
     out_string("`");
     out_string(scontext);
     out_string("'");
 }
-#endif                                /*WITH_SELINUX */
-
 
 static void out_newline(void)
 {
@@ -245,18 +464,15 @@ static PROC *find_proc(pid_t pid)
 {
     PROC *walk;
 
-    for (walk = list; walk; walk = walk->next)
+    for (walk = list; walk; walk = walk->next) {
         if (walk->pid == pid)
-            break;
-    return walk;
+            return walk;
+    }
+    return NULL;
 }
 
-#ifdef WITH_SELINUX
 static PROC *new_proc(const char *comm, pid_t pid, uid_t uid,
                       security_context_t scontext)
-#else                                /*WITH_SELINUX */
-static PROC *new_proc(const char *comm, pid_t pid, uid_t uid)
-#endif                                /*WITH_SELINUX */
 {
     PROC *new;
 
@@ -264,18 +480,18 @@ static PROC *new_proc(const char *comm, pid_t pid, uid_t uid)
         perror("malloc");
         exit(1);
     }
-    strcpy(new->comm, comm);
+    strncpy(new->comm, comm, COMM_LEN+2);
+    new->comm[COMM_LEN+1] = '\0';     /* make sure nul terminated*/
     new->pid = pid;
     new->uid = uid;
     new->flags = 0;
     new->argc = 0;
     new->argv = NULL;
-#ifdef WITH_SELINUX
     new->scontext = scontext;
-#endif                                /*WITH_SELINUX */
     new->children = NULL;
     new->parent = NULL;
     new->next = list;
+    new_proc_ns(new);
     return list = new;
 }
 
@@ -294,8 +510,8 @@ static void add_child(PROC * parent, PROC * child)
         if (by_pid) {
             if ((*walk)->child->pid > child->pid)
                 break;
-        } else if ((cmp = strcmp((*walk)->child->comm, child->comm)) > 0)
-            break;
+        } else if ((cmp = strcmp((*walk)->child->comm, child->comm)) > 0) {
+            break; }
         else if (!cmp && (*walk)->child->uid > child->uid)
             break;
     new->next = *walk;
@@ -333,53 +549,73 @@ static void set_args(PROC * this, const char *args, int size)
         this->argv[i] = start = strchr(start, 0) + 1;
 }
 
-#ifdef WITH_SELINUX
 static void
-add_proc(const char *comm, pid_t pid, pid_t ppid, uid_t uid,
+rename_proc(PROC *this, const char *comm, uid_t uid)
+{
+    PROC *tmp_child, *parent;
+    CHILD **walk;
+
+    strncpy(this->comm, comm, COMM_LEN+2);
+    this->comm[COMM_LEN+1] = '\0';
+    this->uid = uid;
+
+    /* Re-sort children in parent, now we have a name */
+    if (!by_pid && this->parent) {
+        parent = this->parent;
+        for (walk = &parent->children; *walk; walk = &(*walk)->next) {
+            if (
+                ((*walk)->next != NULL) &&
+                strcmp((*walk)->child->comm, (*walk)->next->child->comm) > 0 ) {
+                tmp_child = (*walk)->child;
+                (*walk)->child = (*walk)->next->child;
+                (*walk)->next->child = tmp_child;
+            }
+        }
+    }
+}
+
+static void
+add_proc(const char *comm, pid_t pid, pid_t ppid, pid_t pgid, uid_t uid,
          const char *args, int size, char isthread, security_context_t scontext)
-#else                                /*WITH_SELINUX */
-static void
-add_proc(const char *comm, pid_t pid, pid_t ppid, uid_t uid,
-         const char *args, int size, char isthread)
-#endif                                /*WITH_SELINUX */
 {
     PROC *this, *parent;
 
     if (!(this = find_proc(pid)))
-#ifdef WITH_SELINUX
         this = new_proc(comm, pid, uid, scontext);
-#else                                /*WITH_SELINUX */
-        this = new_proc(comm, pid, uid);
-#endif                                /*WITH_SELINUX */
     else {
-        strcpy(this->comm, comm);
-        this->uid = uid;
+        rename_proc(this, comm, uid);
     }
     if (args)
         set_args(this, args, size);
     if (pid == ppid)
         ppid = 0;
+    this->pgid = pgid;
     if (isthread)
       this->flags |= PFLAG_THREAD;
-    if (!(parent = find_proc(ppid)))
-#ifdef WITH_SELINUX
+    if (!(parent = find_proc(ppid))) {
         parent = new_proc("?", ppid, 0, scontext);
-#else                                /*WITH_SELINUX */
-        parent = new_proc("?", ppid, 0);
-#endif                                /*WITH_SELINUX */
-    add_child(parent, this);
-    this->parent = parent;
+    }
+    if (pid != 0) {
+      add_child(parent, this);
+      this->parent = parent;
+    }
 }
 
 
 static int tree_equal(const PROC * a, const PROC * b)
 {
     const CHILD *walk_a, *walk_b;
+    int i;
 
     if (strcmp(a->comm, b->comm))
         return 0;
     if (user_change && a->uid != b->uid)
         return 0;
+    if (ns_change) {
+        for (i = 0; i < NUM_NS; i++)
+            if (a->ns[i] != b->ns[i])
+                return 0;
+    }
     for (walk_a = a->children, walk_b = b->children; walk_a && walk_b;
          walk_a = walk_a->next, walk_b = walk_b->next)
         if (!tree_equal(walk_a->child, walk_b->child))
@@ -414,7 +650,7 @@ static void
 dump_tree(PROC * current, int level, int rep, int leaf, int last,
           uid_t prev_uid, int closing)
 {
-    CHILD *walk, *next, **scan;
+    CHILD *walk, *next, *tmp_child, **scan;
     const struct passwd *pw;
     int lvl, i, add, offset, len, swapped, info, count, comm_len, first;
     const char *tmp, *here;
@@ -449,6 +685,10 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
         out_char(info++ ? ',' : '(');
         (void) out_int(current->pid);
     }
+    if (pgids) {
+        out_char(info++ ? ',' : '(');
+        (void) out_int(current->pgid);
+    }
     if (user_change && prev_uid != current->uid) {
         out_char(info++ ? ',' : '(');
         if ((pw = getpwuid(current->uid)))
@@ -456,12 +696,20 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
         else
             (void) out_int(current->uid);
     }
-#ifdef WITH_SELINUX
+    if (ns_change && current->parent) {
+        for (i = 0; i < NUM_NS; i++) {
+            if (current->ns[i] == 0 || current->parent->ns[i] == 0)
+                continue;
+            if (current->ns[i] != current->parent->ns[i]) {
+                out_char(info++ ? ',' : '(');
+                out_string(get_ns_name(i));
+            }
+        }
+    }
     if (show_scontext) {
         out_char(info++ ? ',' : '(');
         out_scontext(current->scontext);
     }
-#endif                                /*WITH_SELINUX */
     if ((swapped && print_args && current->argc < 0) || (!swapped && info))
         out_char(')');
     if ((current->flags & PFLAG_HILIGHT) && (tmp = tgetstr("me", NULL)))
@@ -482,11 +730,7 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
             }
         }
     }
-#ifdef WITH_SELINUX
     if (show_scontext || print_args || !current->children)
-#else                                /*WITH_SELINUX */
-    if (print_args || !current->children)
-#endif                                /*WITH_SELINUX */
     {
         while (closing--)
             out_char(']');
@@ -495,11 +739,7 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
     ensure_buffer_capacity(level);
     more[level] = !last;
 
-#ifdef WITH_SELINUX
     if (show_scontext || print_args)
-#else                                /*WITH_SELINUX */
-    if (print_args)
-#endif                                /*WITH_SELINUX */
     {
         width[level] = swapped + (comm_len > 1 ? 0 : -1);
         count=0;
@@ -516,7 +756,9 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
                 if (next == *scan)
                   next = (*scan)->next;
                 count++;
-                *scan = (*scan)->next;
+                tmp_child = (*scan)->next;
+                free(*scan);
+                *scan = tmp_child;
               }
             }
             dump_tree(walk->child, level + 1, count + 1,
@@ -549,7 +791,9 @@ dump_tree(PROC * current, int level, int rep, int leaf, int last,
                     if (next == *scan)
                         next = (*scan)->next;
                     count++;
-                    *scan = (*scan)->next;
+                    tmp_child = (*scan)->next;
+                    free(*scan);
+                    *scan = tmp_child;
                 }
         }
         if (first) {
@@ -581,6 +825,82 @@ static void dump_by_user(PROC * current, uid_t uid)
         dump_by_user(walk->child, uid);
 }
 
+static void dump_by_namespace(struct ns_entry *root)
+{
+    struct ns_entry *ptr = root;
+    CHILD *c;
+    char buff[14];
+
+    for ( ; ptr; ptr = ptr->next) {
+        snprintf(buff, sizeof(buff), "[%li]\n", (long int)ptr->number);
+        out_string(buff);
+        for (c = ptr->children; c; c = c->next)
+            dump_tree(c->child, 0, 1, 1, 1, 0, 0);
+    }
+}
+
+static void trim_tree_by_parent(PROC * current)
+{
+  if (!current)
+    return;
+
+  PROC * parent = current->parent;
+
+  if (!parent)
+    return;
+
+  free_children(parent->children);
+  parent->children = NULL;
+  add_child(parent, current);
+  trim_tree_by_parent(parent);
+}
+
+static char* get_threadname(const pid_t pid, const int tid, const char *comm)
+{
+    FILE *file;
+    char *thread_comm, *endcomm, *threadname;
+    char *path = NULL;
+    size_t len = 0;
+    int nbytes;
+    char readbuf[BUFSIZ + 1];
+
+    if (! (threadname = malloc(COMM_LEN + 2 + 1))) {
+        exit(2);
+    }
+    if (!thread_names) {
+        sprintf(threadname, THREAD_FORMAT, COMM_LEN, comm);
+        return threadname;
+    }
+    len = snprintf(NULL, 0, "%s/%d/task/%d/stat", PROC_BASE, pid, tid);
+    if (len < 0)
+        exit(2);
+    len++;
+    path = malloc(len);
+    if (path == NULL)
+        exit(2);
+    nbytes = snprintf(path, len, "%s/%d/task/%d/stat", PROC_BASE, pid, tid);
+    if (nbytes < 0 || nbytes >= len)
+        perror("get_threadname: snprintf");
+    if ( (file = fopen(path, "r")) != NULL) {
+        if (fgets(readbuf, BUFSIZ, file) != NULL) {
+            if ((thread_comm = strchr(readbuf, '('))
+                && (endcomm = strrchr(thread_comm, ')'))) {
+                ++thread_comm;
+                *endcomm = '\0';
+                sprintf(threadname, THREAD_FORMAT, COMM_LEN, thread_comm);
+                (void) fclose(file);
+                free(path);
+                return threadname;
+            }
+        }
+        fclose(file);
+    }
+    free(path);
+
+    /* Fall back to old method */
+    sprintf(threadname, THREAD_FORMAT, COMM_LEN, comm);
+    return threadname;
+}
 
 /*
  * read_proc now uses a similar method as procps for finding the process
@@ -596,12 +916,12 @@ static void read_proc(void)
   char *buffer;
   size_t buffer_size;
   char readbuf[BUFSIZ + 1];
-  char *tmpptr;
-  pid_t pid, ppid;
+  char *tmpptr, *endptr;
+  pid_t pid, ppid, pgid;
   int fd, size;
   int empty;
-#ifdef WITH_SELINUX
   security_context_t scontext = NULL;
+#ifdef WITH_SELINUX
   int selinux_enabled = is_selinux_enabled() > 0;
 #endif                /*WITH_SELINUX */
 
@@ -621,8 +941,9 @@ static void read_proc(void)
     exit(1);
   }
   empty = 1;
-  while ((de = readdir(dir)) != NULL)
-    if ((pid = (pid_t) atoi(de->d_name)) != 0) {
+  while ((de = readdir(dir)) != NULL) {
+    pid = (pid_t) strtol(de->d_name, &endptr, 10);
+    if (endptr != de->d_name && endptr[0] == '\0') {
       if (! (path = malloc(strlen(PROC_BASE) + strlen(de->d_name) + 10)))
         exit(2);
       sprintf(path, "%s/%d/stat", PROC_BASE, pid);
@@ -652,77 +973,72 @@ static void read_proc(void)
             /* We now have readbuf with pid and cmd, and tmpptr+2
              * with the rest */
             /*printf("tmpptr: %s\n", tmpptr+2); */
-            if (sscanf(tmpptr + 2, "%*c %d", &ppid) == 1) {
+            if (sscanf(tmpptr + 2, "%*c %d %d", &ppid, &pgid) == 2) {
               DIR *taskdir;
               struct dirent *dt;
               char *taskpath;
-              char *threadname;
               int thread;
 
-              if (! (taskpath = malloc(strlen(path) + 10)))
-                exit(2);
-              sprintf(taskpath, "%s/task", path);
+              /* handle process threads */
+              if (! hide_threads) {
+                if (! (taskpath = malloc(strlen(path) + 10)))
+                  exit(2);
+                sprintf(taskpath, "%s/task", path);
 
-              if ((taskdir = opendir(taskpath)) != 0) {
-                /* if we have this dir, we're on 2.6 */
-                if (! (threadname = malloc(COMM_LEN+1))) {
-                    exit(2);
-                }
-                sprintf(threadname, "{%.*s}", COMM_LEN-2, comm);
-                while ((dt = readdir(taskdir)) != NULL) {
-                  if ((thread = atoi(dt->d_name)) != 0) {
-                    if (thread != pid) {
-#ifdef WITH_SELINUX
-                      if (print_args)
-                        add_proc(threadname, thread, pid, st.st_uid, 
-                            threadname, strlen (threadname) + 1, 1,scontext);
-                      else
-                        add_proc(threadname, thread, pid, st.st_uid, 
-                            NULL, 0, 1, scontext);
-#else                /*WITH_SELINUX */
-                      if (print_args)
-                        add_proc(threadname, thread, pid, st.st_uid,
-                            threadname, strlen (threadname) + 1, 1);
-                      else
-                        add_proc(threadname, thread, pid, st.st_uid,
-                            NULL, 0, 1);
-#endif                /*WITH_SELINUX */
+                if ((taskdir = opendir(taskpath)) != 0) {
+                  /* if we have this dir, we're on 2.6 */
+                  while ((dt = readdir(taskdir)) != NULL) {
+                    if ((thread = atoi(dt->d_name)) != 0) {
+                      if (thread != pid) {
+                        char *threadname;
+                        threadname = get_threadname(pid, thread, comm);
+                        if (print_args)
+                          add_proc(threadname, thread, pid, pgid, st.st_uid,
+                              threadname, strlen (threadname) + 1, 1,scontext);
+                        else
+                          add_proc(threadname, thread, pid, pgid, st.st_uid,
+                              NULL, 0, 1, scontext);
+                        free(threadname);
                       }
                     }
                   }
-                  free(threadname);
                   (void) closedir(taskdir);
                 }
-              free(taskpath);
+                free(taskpath);
+              }
+
+              /* handle process */
               if (!print_args)
-#ifdef WITH_SELINUX
-                add_proc(comm, pid, ppid, st.st_uid, NULL, 0, 0, scontext);
-#else                /*WITH_SELINUX */
-                add_proc(comm, pid, ppid, st.st_uid, NULL, 0, 0);
-#endif                /*WITH_SELINUX */
+                add_proc(comm, pid, ppid, pgid, st.st_uid, NULL, 0, 0, scontext);
               else {
                 sprintf(path, "%s/%d/cmdline", PROC_BASE, pid);
                 if ((fd = open(path, O_RDONLY)) < 0) {
-                  perror(path);
-                  exit(1);
+          /* If this fails then the process is gone.  If a PID
+		   * was specified on the command-line then we might
+		   * not even be interested in the current process.
+		   * There's no sensible way of dealing with this race
+		   * so we might as well behave as if the current
+		   * process did not exist.  */
+                    (void) fclose(file);
+                    free(path);
+                    continue;
                 }
                 if ((size = read(fd, buffer, buffer_size)) < 0) {
-                  perror(path);
-                  exit(1);
+                    /* As above. */
+                    close(fd);
+                    (void) fclose(file);
+                    free(path);
+                    continue;
                 }
                 (void) close(fd);
-                /* If we have read the maximum screen length of args, bring it back by one to stop overflow */
-                if (size >= buffer_size)
+                /* If we have read the maximum screen length of args,
+                 * bring it back by one to stop overflow */
+                if (size >= (int)buffer_size)
                   size--;
                 if (size)
                   buffer[size++] = 0;
-#ifdef WITH_SELINUX
-                add_proc(comm, pid, ppid, st.st_uid,
+                add_proc(comm, pid, ppid, pgid, st.st_uid,
                      buffer, size, 0, scontext);
-#else                /*WITH_SELINUX */
-                add_proc(comm, pid, ppid, st.st_uid,
-                     buffer, size, 0);
-#endif                /*WITH_SELINUX */
               }
             }
           }
@@ -731,7 +1047,9 @@ static void read_proc(void)
       }
       free(path);
     }
+  }
   (void) closedir(dir);
+  fix_orphans(scontext);
   if (print_args)
     free(buffer);
   if (empty) {
@@ -741,39 +1059,40 @@ static void read_proc(void)
 }
 
 
-#if 0
-
-/* Could use output of  ps achlx | awk '{ print $3,$4,$2,$13 }'  */
-
-static void read_stdin(void)
+/* When using kernel 3.3 with hidepid feature enabled on /proc
+ * then we need fake root pid and gather all the orphan processes
+ * that is, processes with no known parent
+ * As we cannot be sure if it is just the root pid or others missing
+ * we gather the lot
+ */
+static void fix_orphans(security_context_t scontext)
 {
-    char comm[PATH_MAX + 1];
-    char *cmd;
-    int pid, ppid, uid;
+    PROC *root, *walk;
 
-    while (scanf("%d %d %d %s\n", &pid, &ppid, &uid, comm) == 4) {
-        if (cmd = strrchr(comm, '/'))
-            cmd++;
-        else
-            cmd = comm;
-        if (*cmd == '-')
-            cmd++;
-#ifdef WITH_SELINUX
-        add_proc(cmd, pid, ppid, uid, NULL, 0, NULL);
-#else                                /*WITH_SELINUX */
-        add_proc(cmd, pid, ppid, uid, NULL, 0);
-#endif                                /*WITH_SELINUX */
+    if (!(root = find_proc(ROOT_PID))) {
+        root = new_proc("?", ROOT_PID, 0, scontext);
+    }
+    for (walk = list; walk; walk = walk->next) {
+        if (walk->pid == 1 || walk->pid == 0)
+            continue;
+        if (walk->parent == NULL) {
+            add_child(root, walk);
+            walk->parent = root;
+        }
     }
 }
 
-#endif
-
 
 static void usage(void)
 {
     fprintf(stderr,
             _
-            ("Usage: pstree [ -a ] [ -c ] [ -h | -H PID ] [ -l ] [ -n ] [ -p ] [ -u ]\n"
+            (
+#ifdef WITH_SELINUX
+             "Usage: pstree [-acglpsStuZ] [ -h | -H PID ] [ -n | -N type ]\n"
+#else                                 /*WITH_SELINUX */
+             "Usage: pstree [-acglpsStu] [ -h | -H PID ] [ -n | -N type ]\n"
+#endif                                /*WITH_SELINUX */
              "              [ -A | -G | -U ] [ PID | USER ]\n"
              "       pstree -V\n" "Display a tree of processes.\n\n"
              "  -a, --arguments     show command line arguments\n"
@@ -782,16 +1101,25 @@ static void usage(void)
              "  -h, --highlight-all highlight current process and its ancestors\n"
              "  -H PID,\n"
              "  --highlight-pid=PID highlight this process and its ancestors\n"
+             "  -g, --show-pgids    show process group ids; implies -c\n"
              "  -G, --vt100         use VT100 line drawing characters\n"
              "  -l, --long          don't truncate long lines\n"
              "  -n, --numeric-sort  sort output by PID\n"
+             "  -N type,\n"
+             "  --ns-sort=type      sort by namespace type (cgroup, ipc, mnt, net, pid,\n"
+             "                                              user, uts)\n"
              "  -p, --show-pids     show PIDs; implies -c\n"
+             "  -s, --show-parents  show parents of the selected process\n"
+             "  -S, --ns-changes    show namespace transitions\n"
+             "  -t, --thread-names  show full thread names\n"
+             "  -T, --hide-threads  hide threads, show only processes\n"
              "  -u, --uid-changes   show uid transitions\n"
              "  -U, --unicode       use UTF-8 (Unicode) line drawing characters\n"
              "  -V, --version       display version information\n"));
 #ifdef WITH_SELINUX
     fprintf(stderr,
-            _("  -Z     show         SELinux security contexts\n"));
+            _("  -Z, --security-context\n"
+              "                      show SELinux security contexts\n"));
 #endif                                /*WITH_SELINUX */
     fprintf(stderr, _("  PID    start at this PID; default is 1 (init)\n"
                       "  USER   show only trees rooted at processes of this user\n\n"));
@@ -803,7 +1131,7 @@ void print_version()
     fprintf(stderr, _("pstree (PSmisc) %s\n"), VERSION);
     fprintf(stderr,
             _
-            ("Copyright (C) 1993-2009 Werner Almesberger and Craig Small\n\n"));
+            ("Copyright (C) 1993-2017 Werner Almesberger and Craig Small\n\n"));
     fprintf(stderr,
             _("PSmisc comes with ABSOLUTELY NO WARRANTY.\n"
               "This is free software, and you are welcome to redistribute it under\n"
@@ -815,12 +1143,13 @@ void print_version()
 int main(int argc, char **argv)
 {
     PROC *current;
-    struct winsize winsz;
     const struct passwd *pw;
+    struct ns_entry *nsroot = NULL;
     pid_t pid, highlight;
     char termcap_area[1024];
-    char *termname;
-    int c;
+    char *termname, *endptr;
+    int c, pid_set;
+    enum ns_type nsid = NUM_NS;
 
     struct option options[] = {
         {"arguments", 0, NULL, 'a'},
@@ -829,9 +1158,15 @@ int main(int argc, char **argv)
         {"vt100", 0, NULL, 'G'},
         {"highlight-all", 0, NULL, 'h'},
         {"highlight-pid", 1, NULL, 'H'},
-        {"numeric-sort", 0, NULL, 'n'},
-        {"show-pids", 0, NULL, 'p'},
         {"long", 0, NULL, 'l'},
+        {"numeric-sort", 0, NULL, 'n'},
+        {"ns-sort", 1, NULL, 'N' },
+        {"show-pids", 0, NULL, 'p'},
+        {"show-pgids", 0, NULL, 'g'},
+        {"show-parents", 0, NULL, 's'},
+        {"ns-changes", 0, NULL, 'S' },
+        {"thread-names", 0, NULL, 't'},
+        {"hide-threads", 0, NULL, 'T'},
         {"uid-changes", 0, NULL, 'u'},
         {"unicode", 0, NULL, 'U'},
         {"version", 0, NULL, 'V'},
@@ -841,10 +1176,8 @@ int main(int argc, char **argv)
         { 0, 0, 0, 0 }
     };
 
-    if (ioctl(1, TIOCGWINSZ, &winsz) >= 0)
-        if (winsz.ws_col)
-            output_width = winsz.ws_col;
-    pid = 1;
+    output_width = get_output_width();
+    pid = ROOT_PID;
     highlight = 0;
     pw = NULL;
 
@@ -868,7 +1201,7 @@ int main(int argc, char **argv)
     } else if (isatty(1) && (termname = getenv("TERM")) &&
                (strlen(termname) > 0) &&
                (setupterm(NULL, 1 /* stdout */ , NULL) == OK) &&
-               (tigetstr("acsc") > 0)) {
+               (tigetstr("acsc") != NULL) && (tigetstr("acsc") != (char *)-1)) {
         /*
          * Failing that, if TERM is defined, a non-null value, and the terminal
          * has the VT100 graphics charset, use it.
@@ -884,11 +1217,11 @@ int main(int argc, char **argv)
 
 #ifdef WITH_SELINUX
     while ((c =
-            getopt_long(argc, argv, "aAcGhH:npluUVZ", options,
+            getopt_long(argc, argv, "aAcGhH:nN:pglsStTuUVZ", options,
                         NULL)) != -1)
 #else                                /*WITH_SELINUX */
     while ((c =
-            getopt_long(argc, argv, "aAcGhH:npluUV", options, NULL)) != -1)
+            getopt_long(argc, argv, "aAcGhH:nN:pglsStTuUV", options, NULL)) != -1)
 #endif                                /*WITH_SELINUX */
         switch (c) {
         case 'a':
@@ -930,10 +1263,37 @@ int main(int argc, char **argv)
         case 'n':
             by_pid = 1;
             break;
+        case 'N':
+            nsid = get_ns_id(optarg);
+            if (nsid == NUM_NS)
+                 usage();
+            if (verify_ns(nsid)) {
+                 fprintf(stderr,
+                         _("procfs file for %s namespace not available\n"),
+                         optarg);
+                 return 1;
+            }
+            break;
         case 'p':
             pids = 1;
             compact = 0;
             break;
+        case 'g':
+            pgids = 1;
+            compact = 0;
+            break;
+        case 's':
+            show_parents = 1;
+            break;
+        case 'S':
+            ns_change = 1;
+            break;
+        case 't':
+            thread_names = 1;
+            break;
+        case 'T':
+            hide_threads = 1;
+            break;
         case 'u':
             user_change = 1;
             break;
@@ -957,7 +1317,9 @@ int main(int argc, char **argv)
         }
     if (optind == argc - 1) {
         if (isdigit(*argv[optind])) {
-            if (!(pid = (pid_t) atoi(argv[optind++])))
+            pid = (pid_t) strtol(argv[optind++], &endptr, 10);
+            pid_set = 1;
+            if (endptr[0] != '\0')
                 usage();
         } else if (!(pw = getpwnam(argv[optind++]))) {
             fprintf(stderr, _("No such user name: %s\n"),
@@ -971,16 +1333,28 @@ int main(int argc, char **argv)
     for (current = find_proc(highlight); current;
          current = current->parent)
         current->flags |= PFLAG_HILIGHT;
-    if (!pw)
+
+    if(show_parents && pid_set == 1) {
+      trim_tree_by_parent(find_proc(pid));
+
+      pid = ROOT_PID;
+    }
+
+    if (nsid != NUM_NS) {
+        sort_by_namespace(NULL, nsid, &nsroot);
+        dump_by_namespace(nsroot);
+    } else if (!pw)
         dump_tree(find_proc(pid), 0, 1, 1, 1, 0, 0);
     else {
-        dump_by_user(find_proc(1), pw->pw_uid);
+        dump_by_user(find_proc(ROOT_PID), pw->pw_uid);
         if (!dumped) {
             fprintf(stderr, _("No processes found.\n"));
             return 1;
         }
     }
     free_buffers();
+    free_proc();
+    free_namespace(&nsroot);
     if (wait_end == 1) {
         fprintf(stderr, _("Press return to close\n"));
         (void) getchar();
